.setcpu "65816"
.smart

;Included Files
.include "Header.asm"

;Define Stuff

HVBJOY=		$4212
JOY0=		$4218
JOY1=		$421a
JOY2=		$421c
JOY3=		$421e
JOYSER0=	$4016
JOYSER1=	$4017
WRIO=		$4201

;Button Numbers
RButton=	16
LButton= 	32
XButton= 	64
AButton= 	128

RightDPad=	1
LeftDPad=	2
DownDPad=	4
UpDPad=		8
StartButton=	16
SelectButton=	32
YButton=	64
BButton=	128

WordRightDPad=		256
WordLeftDPad=		512
WordDownDPad=		1024
WordUpDPad=		2048
WordStartButton=	4096
WordSelectButton=	8192
WordYButton=		16384
WordBButton=		32768

NumberOfSlides=		26

;=========================================================================================
.segment "ZEROPAGE"
;=========================================================================================

Joy1Data:			.res 2
Joy2Data:			.res 2
Joy1Press:			.res 2
Joy2Press:			.res 2

BG1X:				.res 2
BG1Y:				.res 2
BG2X:				.res 2
BG2Y:				.res 2
BG3X:				.res 2
BG3Y:				.res 2
BG4X:				.res 2
BG4Y:				.res 2

DisplayController:		.res 2

LongJumpLocation:		.res 3
DMARequestCounter:	.res 2

CurrentSlide:	.res 2
CurrentState:	.res 2
SlideMovementTableOffset:	.res 2

;SNES RAM area, $0200-1fff
;See snes.cfg

;=========================================================================================
.segment "BSS"
;=========================================================================================

SpriteBuf1:			.res $200
SpriteBuf1Size=		*-SpriteBuf1
SpriteBuf2:    		.res $200
SpriteBuf2Size=		*-SpriteBuf2

DMARequestSlotNumber=	128
DMARequestDMAPSettingsTable:	.res 2*DMARequestSlotNumber
DMARequestLoWordTable:		.res 2*DMARequestSlotNumber
DMARequestBankByteTable:	.res 2*DMARequestSlotNumber
DMARequestTransferSizeTable:	.res 2*DMARequestSlotNumber
DMARequestVMAINSettingsTable:	.res 2*DMARequestSlotNumber
DMARequestVramAddressTable:	.res 2*DMARequestSlotNumber
DMARequestCgramAddressTable:	.res 2*DMARequestSlotNumber

;=========================================================================================
.segment "BSS7E"
;=========================================================================================

;=========================================================================================
.segment "CODE"
;=========================================================================================

.proc main

  sei		;Disable interrupts
  clc		;Clear carry, used by XCE
  xce		;Set 65816 native mode
  jml main_fastrom	;Needed to set K (bank of PC) properly if MODE 21 is ever used;
							;see official SNES developers docs, "Programming Cautions"
.endproc		

;=========================================================================================
;=========================================================================================

.proc main_fastrom
  cld		;Disable decimal mode
  phk		;Push K (PC bank)
  plb		;Pull B (data bank, i.e. data bank now equals PC bank)

  rep #$30	;A=16, X/Y=16

;Note: this should correlate with ZEROPAGE in lorom256k.cfg
  lda #$0000
  tcd		;Set D = $0000 (direct page)

;Note: this should correlate with the top of BSS in lorom256k.cfg
  ldx #$1fff
  txs		;Set X = $1fff (stack pointer)

;Register initialisation values, per official Nintendo documentation
;Note: if you are looking at this code and thinking "how can I optimise"
;this?", then you are doing nothing useful with your time.  This code is
;called ONCE on reset or power-on, it does not need to be optimised.  I
;(koitsu) find it remarkable how many people have tried to "optimise"
;this routine.  STOP IT.

  sep #$20	;A=8
  lda #$80
  sta $2100
  stz $2101
  stz $2102
  stz $2103
  stz $2104
  stz $2105
  stz $2106
  stz $2107
  stz $2108
  stz $2109
  stz $210A
  stz $210B
  stz $210C
  stz $210D
  stz $210D
  stz $210E
  stz $210E
  stz $210F
  stz $210F
  stz $2110
  stz $2110
  stz $2111
  stz $2111
  stz $2112
  stz $2112
  stz $2113               
  stz $2113               
  stz $2114
  stz $2114
  lda #$80
  sta $2115
  stz $2116
  stz $2117
  stz $211A
  stz $211B
  lda #$01
  sta $211B
  stz $211C
  stz $211C
  stz $211D
  stz $211D
  stz $211E       
  lda #$01
  sta $211E
  stz $211F
  stz $211F
  stz $2120
  stz $2120
  stz $2121
  stz $2123
  stz $2124
  stz $2125
  stz $2126
  stz $2127
  stz $2128
  stz $2129
  stz $212A
  stz $212B
  stz $212C
  stz $212D
  stz $212E
  stz $212F
  stz $4200
  lda #$FF
  sta $4201
  stz $4202
  stz $4203
  stz $4204
  stz $4205
  stz $4206
  stz $4207
  stz $4208
  stz $4209
  stz $420A
  stz $420B
  stz $420C
  lda #$01
  sta $420D

  jsr clear_vram	;Zero VRAM. Requires A=8, X/Y=16
  jsr clear_cgram	;Zero palette. Requires A=8, X/Y=16

;clear sprite tables

  stz $2102	;sprites initialized to be off the screen, palette 0, character 0
  stz $2103
  ldx #$0080
  lda #$F0

loop08:
  sta $2104	;set X = 240
  sta $2104	;set Y = 240
  sta $2104	;set character = $00
  sta $2104	;set priority=0, no flips
  dex
  bne loop08
  ldx #$0020

loop09:
  stz $2104		;set size bit=0, x MSB = 0
  dex
  bne loop09

;clear SNES RAM

  stz $2181	;set WRAM address to $000000
  stz $2182
  stz $2183

  ldx #$8008
  stx $4300		;Set DMA mode to fixed source, BYTE to $2180
  ldx #.loword(ZeroFillByte)
  stx $4302		;Set source offset
  lda #.bankbyte(ZeroFillByte)
  sta $4304		;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 64KBytes (65536 bytes)
  lda #$01
  sta $420B		;Initiate transfer

  lda #$01		;now zero the next 64KB (i.e. 128KB total)
  sta $420B		;Initiate transfer

  jsr initialize_oam
  jsr initialize_controllers

;=========================================================================================

  stz $211b
  lda #$01
  sta $211b
  stz $211c
  stz $211c
  stz $211d
  stz $211d
  stz $211e
  sta $211e
  lda #$80
  sta $211f
  stz $211f
  sta $2120
  stz $2120

  lda #$09		;Set video mode 1, 16x16 tiles (16 color BG1, 16 color BG2, 4 color BG3)
  sta $2105         

  lda #$0C		;Set BG1's Tile Map VRAM offset to $0000 (word address)
  sta $2107		;and the Tile Map size to 32 tiles x 32 tiles

  lda #$06		;Set BG3's Tile Map VRAM offset to $0000 (word address)
  sta $2109		;and the Tile Map size to 32 tiles x 32 tiles

  lda #$04		;Set BG1's Character VRAM offset to $0000 (word address)
  sta $210B

  lda #$00		;Set BG3's Character VRAM offset to $0000 (word address)
  sta $210C

  lda #$05		;Turn on BG1 and BG3
  sta $212C

  sep #$20
  lda #$00
  sta $2121
  lda #$10
  sta $2122
  lda #$42
  sta $2122

  rep #$20
  lda #$FFFF
  sta BG1Y
  sta BG2Y
  sta BG3Y
  sta BG4Y

  lda #$1801
  sta DMARequestDMAPSettingsTable
  lda #.loword(FontTiles)
  sta DMARequestLoWordTable
  ldx #.bankbyte(FontTiles)
  stx DMARequestBankByteTable
  lda #$0600
  sta DMARequestTransferSizeTable
  ldx #$80
  stx DMARequestVMAINSettingsTable
  lda #$0100
  sta DMARequestVramAddressTable
  stz DMARequestCgramAddressTable

  lda #$1801
  sta DMARequestDMAPSettingsTable+2
  lda #.loword(BackgroundTiles)
  sta DMARequestLoWordTable+2
  ldx #.bankbyte(BackgroundTiles)
  stx DMARequestBankByteTable+2
  lda #$7000
  sta DMARequestTransferSizeTable+2
  ldx #$80
  stx DMARequestVMAINSettingsTable+2
  lda #$4000
  sta DMARequestVramAddressTable+2
  stz DMARequestCgramAddressTable+2

  lda #$1800
  sta DMARequestDMAPSettingsTable+4
  lda #.loword(Slide1Text)
  sta DMARequestLoWordTable+4
  ldx #.bankbyte(Slide1Text)
  stx DMARequestBankByteTable+4
  lda #$0380
  sta DMARequestTransferSizeTable+4
  stz DMARequestVMAINSettingsTable+4
  lda #$0400
  sta DMARequestVramAddressTable+4
  stz DMARequestCgramAddressTable+4

  lda #$1908
  sta DMARequestDMAPSettingsTable+6
  lda #.loword(PriorityByte)
  sta DMARequestLoWordTable+6
  ldx #.bankbyte(PriorityByte)
  stx DMARequestBankByteTable+6
  lda #$0380
  sta DMARequestTransferSizeTable+6
  ldx #$80
  stx DMARequestVMAINSettingsTable+6
  lda #$0400
  sta DMARequestVramAddressTable+6
  stz DMARequestCgramAddressTable+6

  lda #$1908
  sta DMARequestDMAPSettingsTable+8
  lda #.loword(PriorityByte)
  sta DMARequestLoWordTable+8
  ldx #.bankbyte(PriorityByte)
  stx DMARequestBankByteTable+8
  lda #$0380
  sta DMARequestTransferSizeTable+8
  ldx #$80
  stx DMARequestVMAINSettingsTable+8
  lda #$0800
  sta DMARequestVramAddressTable+8
  stz DMARequestCgramAddressTable+8

  lda #$1801
  sta DMARequestDMAPSettingsTable+10
  lda #.loword(BackgroundTilemap)
  sta DMARequestLoWordTable+10
  ldx #.bankbyte(BackgroundTilemap)
  stx DMARequestBankByteTable+10
  lda #$0800
  sta DMARequestTransferSizeTable+10
  ldx #$80
  stx DMARequestVMAINSettingsTable+10
  lda #$0C00
  sta DMARequestVramAddressTable+10
  stz DMARequestCgramAddressTable+10

  lda #$2200
  sta DMARequestDMAPSettingsTable+12
  lda #.loword(BackgroundPalette)
  sta DMARequestLoWordTable+12
  ldx #.bankbyte(BackgroundPalette)
  stx DMARequestBankByteTable+12
  lda #$0020
  sta DMARequestTransferSizeTable+12
  stz DMARequestVMAINSettingsTable+12
  stz DMARequestVramAddressTable+12
  ldx #$10
  stx DMARequestCgramAddressTable+12

  ldx #$0E
  stx DMARequestCounter
  jsr dma_uploader

;Enable NMI (when interrupts get re-enabled) and joypad reading
  sep #$20	;A=8
  lda #$81
  sta $4200
  cli			;Restore interrupts -- we're all ready to go!
  
;=========================================================================================

  wai
  sep #$20
  lda #$00		;Turn on screen, fully dark
  sta DisplayController

;=========================================================================================

fade_in_loop:
  wai
  lda DisplayController
  inc
  sta DisplayController
  cmp #$0F
  bcc fade_in_loop
  rep #$20
  sep #$10

;=========================================================================================

infinite_loop:
  wai			;Wait for interrupts to finish (NMI/VBlank)
  lda CurrentState
  cmp #$0001
  beq move_up_branch
  cmp #$0002
  beq move_down_branch
  bra check_up

move_up_branch:
  brl move_up

move_down_branch:
  brl move_down

check_up:
  lda CurrentSlide
  beq check_down
  lda Joy1Data
  and #WordUpDPad
  beq check_down
  lda #$0001
  sta CurrentState
  dec CurrentSlide
  dec CurrentSlide
  lda #$1800
  sta DMARequestDMAPSettingsTable
  ldx CurrentSlide
  lda SlideAddressTable,x
  sta DMARequestLoWordTable
  lda #.bankbyte(Slide1Text)
  sta DMARequestBankByteTable
  lda #$0380
  sta DMARequestTransferSizeTable
  stz DMARequestVMAINSettingsTable
  lda CurrentSlide
  and #$0002
  bne check_up_odd_numbered_slide
  lda #$0400
  sta DMARequestVramAddressTable
  ldx #$02
  stx DMARequestCounter
  bra infinite_loop

check_up_odd_numbered_slide:
  lda #$0800
  sta DMARequestVramAddressTable
  ldx #$02
  stx DMARequestCounter
  bra infinite_loop

check_down:
  lda CurrentSlide
  cmp #NumberOfSlides
  beq infinite_loop
  lda Joy1Data
  and #WordDownDPad
  beq infinite_loop
  lda #$0002
  sta CurrentState
  inc CurrentSlide
  inc CurrentSlide
  lda #$1800
  sta DMARequestDMAPSettingsTable
  ldx CurrentSlide
  lda SlideAddressTable,x
  sta DMARequestLoWordTable
  lda #.bankbyte(Slide1Text)
  sta DMARequestBankByteTable
  lda #$0380
  sta DMARequestTransferSizeTable
  stz DMARequestVMAINSettingsTable
  lda CurrentSlide
  and #$0002
  bne check_down_odd_numbered_slide
  lda #$0400
  sta DMARequestVramAddressTable
  ldx #$02
  stx DMARequestCounter
  brl infinite_loop

check_down_odd_numbered_slide:
  lda #$0800
  sta DMARequestVramAddressTable
  ldx #$02
  stx DMARequestCounter
infinite_loop_branch:
  brl infinite_loop

;=========================================================================================

move_up:
  ldx SlideMovementTableOffset
  lda BG3Y
  sec
  sbc a:SlideMovementTable,x
  sta BG3Y
  inx
  inx
  stx SlideMovementTableOffset
  cpx #$3E
  bne infinite_loop_branch
  stz CurrentState
  stz SlideMovementTableOffset
  brl infinite_loop

move_down:
  ldx SlideMovementTableOffset
  lda BG3Y
  clc
  adc a:SlideMovementTable,x
  sta BG3Y
  inx
  inx
  stx SlideMovementTableOffset
  cpx #$3E
  bne infinite_loop_branch
  stz CurrentState
  stz SlideMovementTableOffset
  brl infinite_loop

.endproc

;=========================================================================================
;=========================================================================================

.proc vblank
  jml vblank_fastrom
.endproc

;=========================================================================================
;=========================================================================================

.proc vblank_fastrom
  rep #$30	;A=16, X/Y=16
  pha
  phx
  phy
  phd
  phb
  php

  rep #$20
  lda DisplayController
  sta $2100
  
  sep #$20
  lda BG1X
  sta a:$210D
  lda BG1X+1
  sta a:$210D
  lda BG1Y
  sta a:$210E
  lda BG1Y+1
  sta a:$210E
  lda BG2X
  sta a:$210F
  lda BG2X+1
  sta a:$210F
  lda BG2Y
  sta a:$2110
  lda BG2Y+1
  sta a:$2110
  lda BG3X
  sta a:$2111
  lda BG3X+1
  sta a:$2111
  lda BG3Y
  sta a:$2112
  lda BG3Y+1
  sta a:$2112
  lda BG4X
  sta a:$2113
  lda BG4X+1
  sta a:$2113
  lda BG4Y
  sta a:$2114
  lda BG4Y+1
  sta a:$2114

  jsr dma_uploader
  jsr get_input	;update joypad data

  lda $4210	;clear NMI Flag

  rep #$30
  plp
  plb
  pld
  ply
  plx
  pla
  rti
.endproc

;=========================================================================================
;=========================================================================================

.proc clear_vram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  lda #$80
  sta $2115		;Set VRAM port to word access
  ldx #$1809
  stx $4300		;Set DMA mode to fixed source, WORD to $2118/9
  ldx #$0000
  stx $2116			;Set VRAM port address to $0000
  ldx #.loword(ZeroFillByte)
  stx $4302			;Set source address to $xx:0000
  lda #.bankbyte(ZeroFillByte)
  sta $4304			;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 65536 bytes
  lda #$01
  sta $420B		;Initiate transfer
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc clear_cgram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  stz $2121
  ldx #$0100

clear_palette_loop:
  stz $2122
  stz $2122
  dex
  bne clear_palette_loop
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc initialize_oam
  php	
  rep #$30

  lda #$01
  ldx #0

initialize_oam_loop:
  sta SpriteBuf2+1,x
  inx
  inx
  inx
  inx
  cpx #$0200
  bne initialize_oam_loop
  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc initialize_controllers
  php	

  rep	#$10	;8 bit mem/A, 16 bit X/Y
  sep	#$20

  lda #$C0	;have the automatic read of the SNES read the first pair of JoyPads
  sta WRIO

  ldx #$0000
  stx Joy1Press
  stx Joy2Press

;in ZSNES, the auto-read doesn't work unless NMI is enabled... is this correct?
;
; lda $4212
; lda $4212
;
;w11:
; lda $4212
; and #$01
; beq _w11	;Wait till automatic JoyPort read starts
;
; jsr GetInput	;fill the variables with real Joypad data
;
;w12:
; lda $4212
; and #$01
; beq _w12	;Wait till automatic JoyPort read starts
;
; jsr get_input	;fill the variables with real Joypad data

  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc get_input
  php
  sep #$20	;A=8

  lda #$01
: bit $4212
  bne :-		;Wait till automatic JoyPort read is complete

  rep #$30   ;A=16, X/Y=16

;Get Joypads 1,2

  lda JOY0		;get JoyPad1
  tax 			
  eor Joy1Data		;A = A xor JoyState = (changes in joy state)
  stx Joy1Data		;update JoyState
  ora Joy1Press		;A = (joy changes) or (buttons pressed)
  and Joy1Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy1Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

  lda JOY1		;get JoyPad2
  tax 			
  eor Joy2Data		;A = A xor JoyState = (changes in joy state)
  stx Joy2Data		;update JoyState
  ora Joy2Press		;A = (joy changes) or (buttons pressed)
  and Joy2Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy2Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

;Make sure Joypads 1,2 are valid

  sep #$30	;A=8, X/Y=8
  lda JOYSER0
  eor #$01
  and #$01		;A = -bit0 of JoySer0
  ora Joy1Data
  sta Joy1Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

  lda JOYSER1
  eor #$01
  and #$01		;A = -bit0 of JoySer1
  ora Joy2Data		
  sta Joy2Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

;Change all invalid joypads to have a state of no button presses

  rep #$30	;A=16, X/Y=16

  ldx #$0001
  lda #$000f

  bit Joy1Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq :+		;into the joy port, or it is not a joypad
  stx Joy1Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy1Press

: bit Joy2Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq done		;into the joy port, or it is not a joypad
  stx Joy2Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy2Press

done:
  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc dma_uploader
  sep #$10
  rep #$20
  lda #$4300
  tcd
  lda a:DMARequestCounter
  beq dma_uploader_done
  ldx #$00

dma_uploader_loop:
  lda a:DMARequestDMAPSettingsTable,x
  sta $00

  lda a:DMARequestLoWordTable,x
  sta $02

  ldy a:DMARequestBankByteTable,x
  sty $04

  lda a:DMARequestTransferSizeTable,x
  sta $05

  lda a:DMARequestVMAINSettingsTable,x
  sta a:$2115

  lda a:DMARequestVramAddressTable,x
  sta a:$2116

  ldy a:DMARequestCgramAddressTable,x
  sty a:$2121

  ldy #$01		;Initiate DMA transfer (channel 0)
  sty a:$420B

  inx
  inx
  beq dma_uploader_done
  cpx a:DMARequestCounter
  bne dma_uploader_loop

dma_uploader_done:
  lda #$0000
  tcd
  stz DMARequestCounter
  rts
.endproc

;=========================================================================================
.segment "RODATA"
;=========================================================================================

SlideMovementTable:
  .word $0001,$0002,$0003,$0004,$0005,$0006,$0007,$0008,$0009,$000A,$000B,$000C,$000D,$000E,$000F,$0010,$000F,$000E,$000D,$000C,$000B,$000A,$0009,$0008,$0007,$0006,$0005,$0004,$0003,$0002,$0001

SlideAddressTable:
  .word .loword(Slide1Text),.loword(Slide2Text),.loword(Slide3Text),.loword(Slide4Text),.loword(Slide5Text),.loword(Slide6Text),.loword(Slide7Text)
  .word .loword(Slide8Text),.loword(Slide9Text),.loword(Slide10Text),.loword(Slide11Text),.loword(Slide12Text),.loword(Slide13Text),.loword(Slide14Text)

;=========================================================================================

PriorityByte:
  .byte $20

Slide1Text:
  .incbin "Slide1.bin"

Slide2Text:
  .incbin "Slide2.bin"

Slide3Text:
  .incbin "Slide3.bin"

Slide4Text:
  .incbin "Slide4.bin"

Slide5Text:
  .incbin "Slide5.bin"

Slide6Text:
  .incbin "Slide6.bin"

Slide7Text:
  .incbin "Slide7.bin"

Slide8Text:
  .incbin "Slide8.bin"

Slide9Text:
  .incbin "Slide9.bin"

Slide10Text:
  .incbin "Slide10.bin"

Slide11Text:
  .incbin "Slide11.bin"

Slide12Text:
  .incbin "Slide12.bin"

Slide13Text:
  .incbin "Slide13.bin"

Slide14Text:
  .incbin "Slide14.bin"

;=========================================================================================

BackgroundTilemap:
  .incbin "Background.map"

;=========================================================================================

BackgroundPalette:
  .incbin "Background.clr"

;=========================================================================================

;Byte used to zero/initialise all of memory.
;This is needed because DMA requires an address of a byte to
;read from, not a raw value.

ZeroFillByte:
  .byte $00

;=========================================================================================
.segment "RODATA2"
;=========================================================================================

BackgroundTiles:
  .incbin "Background.pic"
  
FontTiles:
  .incbin "Font.pic"

;=========================================================================================